
package fr.inria.tea.aadl2ssme.transformation.popup.actions;

import java.util.ArrayList;
import org.osate.aadl2.*;
import org.eclipse.pop.ssme.*;
import org.eclipse.pop.ssme.impl.*;

public class ASME2SSME_Translate_DataPort extends ASME2SSME_Translate_Port {
	
	private signalTreeAPI myLib;
	private AADLCoLAPI aadlColAPI;

	/**
	 * Class constructor
	 */
	public ASME2SSME_Translate_DataPort() {
		myLib= new signalTreeAPI();
		aadlColAPI = new AADLCoLAPI();
	}

	/** 
	 * Returns a Signal process that represents an AADL in data port of a thread
	 * <p>
	 * This Signal process is xx_InDataPort, where xx is the name of the data port.  
	 * It is composed of two subprocesses: xx_InDataPort_Behavior() and xx_InDataPort_Property(). 
	 * xx_InDataPort_Behavior() calls the InDataPort_Behavior() process from the predefined library. 
	 * xx_InDataPort_Property() calls the processes that represent the properties.
	 * 
	 * @param timpl thread implementation component
	 * @param dataPort an AADL data port of thread timpl
	 * @return a Signal process
	 */	
	public ProcessModelImpl ASME2SSME_Thread_InDataPort(ThreadImplementation timpl, DataPort dataPort){		

		String name = aadlColAPI.Port_GetName(dataPort);
		String pname = Name_of_ThreadDataPort(name, "in");

		Trace("\n ######## ASME2SSME_Translate_DataPort: ASME2SSME_Thread_InDataPort : " + name);			
		ProcessModelImpl vprocess = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		//parameters
		// type item_type;
		TypesDeclarationImpl item_type = new TypesDeclarationImpl();
		item_type.getTypeDefinition().add(myLib.mkIdentifier(ITEM_TYPE_ID));

		//add parameters to interface
		ArrayList<ExprParam> lparams = new ArrayList<ExprParam>();
		lparams.add(myLib.mkIdentifier(ITEM_TYPE_ID));
		myLib.STree_addDeclarationToInterface(vprocess, item_type, 1);

		//inputs
		// event Frozen_time_event;
		SignalsDeclarationImpl Frozen_time_event = myLib.mkEventDecl(FROZEN_TIME_EVENT_ID);
		//item_type write_flow;
		SignalsDeclarationImpl write_flow = myLib.mkDecl( WRITE_FLOW_ID, ITEM_TYPE_ID  );
		//event Reference_time_event;
		SignalsDeclarationImpl Reference_time_event = myLib.mkEventDecl(REFERENCE_TIME_EVENT_ID);
		// add inputs to interface
		ArrayList<ExprSignal> lins = new ArrayList<ExprSignal>();
		lins.add(myLib.mkIdentifier(FROZEN_TIME_EVENT_ID));
		lins.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		lins.add(myLib.mkIdentifier(REFERENCE_TIME_EVENT_ID));
		myLib.STree_addDeclarationToInterface(vprocess, Frozen_time_event, 2);
		myLib.STree_addDeclarationToInterface(vprocess, write_flow, 2);
		myLib.STree_addDeclarationToInterface(vprocess, Reference_time_event, 2);

		//outputs
		// item_type read_flow;
		SignalsDeclarationImpl read_flow = myLib.mkDecl(READ_FLOW_ID, ITEM_TYPE_ID ); 

		// add outputs to interface
		ArrayList<ExprSignal> louts = new ArrayList<ExprSignal>();
		louts.add(myLib.mkIdentifier(READ_FLOW_ID));
		myLib.STree_addDeclarationToInterface(vprocess, read_flow, 3);

		//in data port behavior
		ProcessModelImpl t_behavior = ASME2SSME_Thread_InDataPort_behavior(pname);
		String threadindataport_behavior_name = myLib.STree_GetStringName(t_behavior.getIdentifier());

		//add to declaration
		myLib.STree_addProcessDeclaration(vprocess, t_behavior);
		//add to body
		IdentifierImpl id = (IdentifierImpl) myLib.mkIdentifier(threadindataport_behavior_name);	
		ArrayList<ExprSignal> lins2 = new ArrayList<ExprSignal>();
		lins2.add(myLib.mkIdentifier(FROZEN_TIME_EVENT_ID));
		lins2.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		SignalsDefinitionImpl eq_behavior_call = myLib.STree_mkCompleteProcessCallEquation(id, lparams, lins2, louts);
		myLib.STree_addExpressionOfProcessToaModel(vprocess, eq_behavior_call);
		myLib.STree_addComment(eq_behavior_call , "in data port behavior sub process", false);	
		//in data port property
		ProcessModelImpl t_property = ASME2SSME_Thread_Port_property(timpl, dataPort);

		//add fix number in data port properties
		ASME2SSME_Thread_DataPort_FixProperties(timpl, dataPort, t_property);

		String property_name = myLib.STree_GetStringName(t_property.getIdentifier());
		//add to declaration
		myLib.STree_addProcessDeclaration(vprocess, t_property);
		//add to body
		ArrayList<ExprParam> lparam3 = new ArrayList<ExprParam>();
		ArrayList<ExprSignal> lin3 = new ArrayList<ExprSignal>();
		lin3.add(myLib.mkIdentifier(FROZEN_TIME_EVENT_ID));
		lin3.add(myLib.mkIdentifier(REFERENCE_TIME_EVENT_ID));
		ProcessCallImpl property_inst = myLib.STree_mkCompleteProcessCall(myLib.mkIdentifier(property_name),lparam3, lin3);
		myLib.STree_addExpressionOfProcessToaModel(vprocess, property_inst);
		myLib.STree_addComment(property_inst, "in data port property sub process", false);
		Trace("\n ######################### end ASME2SSME_Thread_InDataPort : " + name + "######################### \n");
		return vprocess;
	}

	/**
	 * Returns a Signal process "xx_InDataPort_behavior", which is a subprocess of thread in data port
	 * <p>
	 * This Signal process is named as xx_InDataPort_behavior, where xx is the data port name. 
	 * The interface contains parameters: {type item_type;} 
	 *<p>
	 * inputs: ? event Frozen_time_event; item_type write_flow; event Reference_time_event; 
	 * outputs: ! item_type read_flow; 
	 * The body calls the InDataPort_behavior() process which is defined in our Signal library. 
	 * 
	 * @param t_name a String name of thread in data port Signal process
	 * @return a Signal process
	 */
	public ProcessModelImpl ASME2SSME_Thread_InDataPort_behavior(String t_name){

		//		* process xx_InDataPort_behavior =
		//			 * {type item_type;
		//			 *  item_type init_value;}
		//			 * (? event Frozen_time_event;
		//			 *    item_type write_flow;
		//			 *    event Reference_time_event;
		//			 *  ! item_type read_flow;)
		//			 *  (|read_flow := InDataPort_behavior{item_type,init_value}(Frozen_time_event,write_flow,Reference_time_event) 
		//		         |)

		Trace("\n ######## ASME2SSME_Translate_DataPort: ASME2SSME_Thread_InDataPort_behavior : " + t_name);

		String pname =  t_name + "_behavior";
		ProcessModelImpl t_behavior = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		//parameters
		// type item_type;
		TypesDeclarationImpl item_type = new TypesDeclarationImpl();
		item_type.getTypeDefinition().add(myLib.mkIdentifier(ITEM_TYPE_ID));

		//add parameters to interface
		ArrayList<ExprParam> lparams = new ArrayList<ExprParam>();
		lparams.add(myLib.mkIdentifier(ITEM_TYPE_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, item_type, 1);

		//inputs
		// event Frozen_time_event;
		SignalsDeclarationImpl Frozen_time_event = myLib.mkEventDecl (FROZEN_TIME_EVENT_ID);
		//item_type write_flow;
		SignalsDeclarationImpl write_flow = myLib.mkDecl(WRITE_FLOW_ID, ITEM_TYPE_ID );

		// add inputs to interface
		ArrayList<ExprSignal> lins = new ArrayList<ExprSignal>();
		lins.add(myLib.mkIdentifier(FROZEN_TIME_EVENT_ID));
		lins.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, Frozen_time_event, 2);
		myLib.STree_addDeclarationToInterface(t_behavior, write_flow, 2);

		//outputs
		// item_type read_flow;
		SignalsDeclarationImpl read_flow =myLib.mkDecl(READ_FLOW_ID, ITEM_TYPE_ID );

		// add outputs to interface
		ArrayList<ExprSignal> louts = new ArrayList<ExprSignal>();
		louts.add(myLib.mkIdentifier(READ_FLOW_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, read_flow, 3);

		//body
		SignalsDefinitionImpl eq1 = myLib.STree_mkCompleteProcessCallEquation(myLib.mkIdentifier("InDataPort_behavior"), lparams, lins, louts);
		myLib.STree_addExpressionOfProcessToaModel(t_behavior, eq1);

		//Trace(t_behavior.getIdentifier().getAtomString());
		return t_behavior;
	}


	/** 
	 * Returns a Signal process xx_OutDataPort() that represents an out data port of a thread
	 * <p>
	 * This process is named as "xx_OutDataPort", where xx is the name of this data port. 
	 * The interface contains the parameters: type item_type; 
	 * inputs: event Output_time_event; item_type write_flow; event Reference_time_event;  
	 * output: item_type sent_flow;
	 * <p>
	 * The body is composed of two subprocesses: xx_OutDataPort_Behavior() and xx_OutDataPort_Property(). 
	 * xx_OutDataPort_Behavior() calls the OutDataPort_Behavior() process from the library. 
	 * xx_OutDataPort_Property() calls properties of this data port.
	 * 
	 * @param timpl a thread implementation component
	 * @param dataPort an out data port of this thread
	 * @return a Signal process
	 */	
	public ProcessModelImpl ASME2SSME_Thread_OutDataPort(ThreadImplementation timpl, DataPort dataPort){		

		Trace("\n ######## ASME2SSME_Translate_DataPort: ASME2SSME_Thread_OutDataPort " + dataPort.getName());

		String name = aadlColAPI.Port_GetName(dataPort);
		String pname = Name_of_ThreadDataPort(name, "out");			
		ProcessModelImpl vprocess = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		//parameters
		// type item_type;
		TypesDeclarationImpl item_type = new TypesDeclarationImpl();
		item_type.getTypeDefinition().add(myLib.mkIdentifier(ITEM_TYPE_ID));

		//add parameters to interface
		ArrayList<ExprParam> lparams = new ArrayList<ExprParam>();
		lparams.add(myLib.mkIdentifier(ITEM_TYPE_ID));
		myLib.STree_addDeclarationToInterface(vprocess, item_type, 1);

		//inputs
		// event Frozen_time_event;
		SignalsDeclarationImpl Output_time_event = myLib.mkEventDecl(OUTPUT_TIME_EVENT_ID ); 
		//item_type write_flow;
		SignalsDeclarationImpl write_flow = myLib.mkDecl(WRITE_FLOW_ID, ITEM_TYPE_ID);
		//event Reference_time_event;
		SignalsDeclarationImpl Reference_time_event =myLib.mkEventDecl(REFERENCE_TIME_EVENT_ID);
		// add inputs to interface
		ArrayList<ExprSignal> lins = new ArrayList<ExprSignal>();
		lins.add(myLib.mkIdentifier(OUTPUT_TIME_EVENT_ID));
		lins.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		lins.add(myLib.mkIdentifier(REFERENCE_TIME_EVENT_ID));
		myLib.STree_addDeclarationToInterface(vprocess, Output_time_event, 2);
		myLib.STree_addDeclarationToInterface(vprocess, write_flow, 2);
		myLib.STree_addDeclarationToInterface(vprocess, Reference_time_event, 2);
		//outputs
		// item_type read_flow;
		SignalsDeclarationImpl sent_flow = myLib.mkDecl(SENT_FLOW_ID, ITEM_TYPE_ID);
		// add outputs to interface
		ArrayList<ExprSignal> louts = new ArrayList<ExprSignal>();
		louts.add(myLib.mkIdentifier(SENT_FLOW_ID));
		myLib.STree_addDeclarationToInterface(vprocess, sent_flow, 3);
		//out data port behavior
		ProcessModelImpl t_behavior = ASME2SSME_Thread_OutDataPort_behavior(pname);
		String threadoutdataport_behavior_name = myLib.STree_GetStringName(t_behavior.getIdentifier());//pname + "_behavior";
		//add to declaration
		myLib.STree_addProcessDeclaration(vprocess, t_behavior);
		//add to body
		IdentifierImpl id = (IdentifierImpl) myLib.mkIdentifier(threadoutdataport_behavior_name);
		ArrayList<ExprSignal> lins2 = new ArrayList<ExprSignal>();
		lins2.add(myLib.mkIdentifier(OUTPUT_TIME_EVENT_ID));
		lins2.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		//lins2.remove(myLib.mkIdentifier(REFERENCE_TIME_EVENT_ID));
		SignalsDefinitionImpl eq_behavior_call = myLib.STree_mkCompleteProcessCallEquation(id, lparams, lins2, louts);
		myLib.STree_addExpressionOfProcessToaModel(vprocess, eq_behavior_call);
		myLib.STree_addComment(eq_behavior_call, "out data port behavior sub process"  , false);

		//in data port property
		ProcessModelImpl t_property = ASME2SSME_Thread_Port_property(timpl,dataPort);

		//add fix number out data port properties
		ASME2SSME_Thread_DataPort_FixProperties(timpl, dataPort, t_property);

		String property_name = myLib.STree_GetStringName(t_property.getIdentifier());//pname + "_behavior";
		//add to declaration
		myLib.STree_addProcessDeclaration(vprocess, t_property);
		//add to body
		ArrayList<ExprParam> lparam3 = new ArrayList<ExprParam>();
		ArrayList<ExprSignal> lin3 = new ArrayList<ExprSignal>();
		lin3.add(myLib.mkIdentifier(OUTPUT_TIME_EVENT_ID));
		lin3.add(myLib.mkIdentifier(REFERENCE_TIME_EVENT_ID));
		ProcessCallImpl property_inst = myLib.STree_mkCompleteProcessCall(myLib.mkIdentifier(property_name),lparam3, lin3);
		myLib.STree_addExpressionOfProcessToaModel(vprocess, property_inst);
		myLib.STree_addComment(property_inst, "out data port property sub process", false);

		Trace("\n ######################### end ASME2SSME_Thread_OutDataPort : " + dataPort.getName() + "######################### \n");
		return vprocess;
	}


	/** 
	 * Returns a Signal process xx_OutDataPort_behavior() that represents the behavior of an out data port of a thread
	 * <p>
	 * This process is named as "xx_OutDataPort_behavior", where xx is the name of this data port. 
	 * The interface contains the parameters: type item_type; 
	 * inputs: event Output_time_event; item_type write_flow; event Reference_time_event;  
	 * output: item_type sent_flow;
	 * <p>
	 * The body calls OutDtaPort_behavior() process which is a predefined process in the library.
	 * 
	 * @param t_name a String of Signal process name of this thread out data port
	 * @return a Signal process
	 */
	public ProcessModelImpl ASME2SSME_Thread_OutDataPort_behavior(String t_name){

		Trace("\n ######## ASME2SSME_Translate_DataPort: ASME2SSME_Thread_OutDataPort_behavior : " + t_name);

		String pname =  t_name + "_behavior";
		ProcessModelImpl t_behavior = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		//parameters
		// type item_type;
		TypesDeclarationImpl item_type = new TypesDeclarationImpl();
		item_type.getTypeDefinition().add(myLib.mkIdentifier(ITEM_TYPE_ID));

		//add parameters to interface
		ArrayList<ExprParam> lparams = new ArrayList<ExprParam>();
		lparams.add(myLib.mkIdentifier(ITEM_TYPE_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, item_type, 1);

		//inputs
		// event Output_time_event;
		SignalsDeclarationImpl Output_time_event = myLib.mkEventDecl(OUTPUT_TIME_EVENT_ID);
		//item_type write_flow;
		SignalsDeclarationImpl write_flow = myLib.mkDecl(WRITE_FLOW_ID, ITEM_TYPE_ID );

		// add inputs to interface
		ArrayList<ExprSignal> lins = new ArrayList<ExprSignal>();
		lins.add(myLib.mkIdentifier(OUTPUT_TIME_EVENT_ID));
		lins.add(myLib.mkIdentifier(WRITE_FLOW_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, Output_time_event, 2);
		myLib.STree_addDeclarationToInterface(t_behavior, write_flow, 2);

		//outputs
		// item_type sent_flow;
		SignalsDeclarationImpl sent_flow = myLib.mkDecl(SENT_FLOW_ID, ITEM_TYPE_ID);

		// add outputs to interface
		ArrayList<ExprSignal> louts = new ArrayList<ExprSignal>();
		louts.add(myLib.mkIdentifier(SENT_FLOW_ID));
		myLib.STree_addDeclarationToInterface(t_behavior, sent_flow, 3);

		//body
		SignalsDefinitionImpl eq1 = myLib.STree_mkCompleteProcessCallEquation(myLib.mkIdentifier("OutDataPort_behavior"), lparams, lins, louts);
		myLib.STree_addExpressionOfProcessToaModel(t_behavior, eq1);

		//Trace(t_behavior.getIdentifier().getAtomString());
		return t_behavior;
	}


	/**
	 * Translates some properties which are fixed for data port
	 * 
	 * @param timpl
	 * @param p
	 * @param t_property
	 */
	public void ASME2SSME_Thread_DataPort_FixProperties(ThreadImplementation timpl, DataPort p, ProcessModelImpl t_property){

		Trace("\n ######## ASME2SSME_Translate_DataPort: ASME2SSME_Thread_DataPort_FixProperties " + p.getName());
		if (aadlColAPI.isInputPort(p)) {
			//Input_Time property
			myLib.STree_addExpressionOfProcessToaModel(t_property,   ASME2SSME_InputTime(timpl, p));
		}
		else if(aadlColAPI.isOutputPort(p)){
			//Input_Time property
			myLib.STree_addExpressionOfProcessToaModel(t_property, ASME2SSME_OutputTime(timpl, p));
			//Timing property
			myLib.STree_addExpressionOfProcessToaModel(t_property, ASME2SSME_Timing(timpl, p));
		}
		else 
			Trace("Port direction out of consideration");
	}
}
