 
package fr.inria.tea.aadl2ssme.transformation.popup.actions;

import java.util.ArrayList;
import org.osate.aadl2.*;

import org.eclipse.pop.ssme.*;
import org.eclipse.pop.ssme.impl.*;

public class ASME2SSME_Translate_Device extends ASME2SSME_Translate_Feature{
	private AADLCoLAPI aadlColAPI;
	private signalTreeAPI myLib;
	private ASME2SSME_Translate_Common asme2ssme_common;
	/**
	 * Class constructor
	 */
	public ASME2SSME_Translate_Device() {
		aadlColAPI = new AADLCoLAPI();
		myLib= new signalTreeAPI();
		asme2ssme_common = new ASME2SSME_Translate_Common();
	}

	/**
	 * Returns a Signal process that represents a device type component
	 * <p>
	 * Gets the in/out port of this device type, and translates them as the input/output of the interface (use ASME2SSME_ComponentInterface())
	 * 
	 * @param d a device type component
	 * @return a Signal process
	 */
	public ProcessModelImpl ASME2SSME_DeviceType(DeviceType d){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_DeviceType: "+ d.getName());
		String sname = aadlColAPI.ComponentType_GetName(d);			
		String pname = Name_of_Device(sname);
		ProcessModelImpl dprocess = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );
		// process interface
		asme2ssme_common.ASME2SSME_ComponentInterface(dprocess, d);
		//Features 
		ASME2SSME_Device_Features(d, dprocess);
		//Flow specifications 
		ASME2SSME_Device_FlowSpecs(dprocess, d);
		//Modes : not exist in version 1.5.8
		//Properties: properties implemented in the DeviceImpl_Property() sub process
		return dprocess;		
	}	


	/** 
	 * Returns a Signal process that represents a device implementation.
	 * <p>
	 * Gets its device type, and translates it as the interface of this process. 
	 * Gets its subcomponents, and each subcomponent classifier (type or implementation) is translated by the corresponding rule. 
	 * (They are represented as instances of their classifiers. )
	 * 
	 * @param d a device implementation component
	 * @return a ProcessModelImpl object (Signal process)	
	 */
	public ProcessModelImpl ASME2SSME_DeviceImpl(DeviceImplementation d){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_DeviceImpl: "+ d.getName());
		String sname = aadlColAPI.ComponentImpl_GetName(d);			
		String pname = Name_of_Device(sname);
		// device process skeleton
		ProcessModelImpl dprocess = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		// device process interface 
		DeviceType device_type = (DeviceType) aadlColAPI.ComponentImpl_GetComponentType(d);
		asme2ssme_common.ASME2SSME_ComponentInterface(dprocess, device_type);

		//Features 
		ASME2SSME_Device_Features(device_type, dprocess);

		//device behavior
		ProcessModelImpl device_behavior = ASME2SSME_Device_behavior(d);
		String device_behavior_name = myLib.STree_GetStringName(device_behavior.getIdentifier()); 
		//add to declaration
		myLib.STree_addProcessDeclaration(dprocess, device_behavior);
		//add to body
		//ProcessCallImpl device_behavior_inst = myLib.STree_mkModelCall(device_behavior_name);
		ArrayList<ExprParam> lparam = new ArrayList<ExprParam>();
		ProcessExpansionImpl device_behavior_inst = myLib.STree_mkProcessCallWithParameters(myLib.mkIdentifier(device_behavior_name), lparam);
		myLib.STree_addExpressionOfProcessToaModel(dprocess, device_behavior_inst);

		//add comments 
		ArrayList<String> comments1 = new ArrayList<String>();
		comments1.add("device behavior sub process");
		myLib.STree_addComment(device_behavior_inst, comments1, false);

		//process property 
		ProcessModelImpl device_property = ASME2SSME_Device_property(d);  
		String device_property_name = myLib.STree_GetStringName(device_property.getIdentifier()); 
		//add to declaration
		myLib.STree_addProcessDeclaration(dprocess, device_property);
		//add to body
		ProcessCallImpl device_property_inst = myLib.STree_mkModelCall(device_property_name);
		myLib.STree_addExpressionOfProcessToaModel(dprocess, device_property_inst);

		//add comments 
		ArrayList<String> comments2 = new ArrayList<String>();
		comments2.add("device property sub process");
		myLib.STree_addComment(device_property_inst, comments2, false);

		//Flow specifications
		ASME2SSME_Device_FlowSpecs(dprocess, device_type);

		//Modes: no modes in version 1.5.8

		//Properties: properties implemented in the DeviceImpl_Property() sub process

		//Subcomponents: no subcomponents exist in version 1.5.8

		//Connections: no connections exist in version 1.5.8c

		//Flows 
		ASME2SSME_DeviceImpl_Flows(dprocess, d);
		//Modes 
		ASME2SSME_DeviceImpl_Modes(d);
		return dprocess;
	}


	/** 
	 * Returns a Signal process "xx_Device_behavior" that represents the behavior of a device component
	 * <p>
	 * This process name is "xxx_behavior", where xxx is this device's corresponding Signal process name. 
	 * In this moment, only a skeleton process model is provided. 
	 * The interface and body are empry.
	 * 
	 * @param p_name a string name of a Signal process which represents this device component
	 * @return a ProcessModelImpl object (a Signal process)
	 */
	public ProcessModelImpl ASME2SSME_Device_behavior(DeviceImplementation d){

		String sname = aadlColAPI.ComponentImpl_GetName(d);	
		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_Device_behavior: "+ sname);

		String bname = Name_of_Device(sname) + "_behavior";
		ProcessModelImpl p_behavior = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(bname) );

		// device behavior process interface
		DeviceType device_type = (DeviceType) aadlColAPI.ComponentImpl_GetComponentType(d);
		asme2ssme_common.ASME2SSME_ComponentInterface(p_behavior, device_type);

		//body: external  
		DescriptionImpl  desc = (DescriptionImpl) p_behavior.getDescription();	
		if (desc == null) {
			p_behavior.setDescription(new DescriptionImpl());
			desc = (DescriptionImpl) p_behavior.getDescription();
		}
		ExternalAttributeImpl externe = new ExternalAttributeImpl();
		desc.setBody(externe);
		return p_behavior;
	}


	/**
	 * Returns a Signal process xx_device_property that represents the properties of a device component.
	 * <p>
	 * This process is named as xxx_property, where xxx is the name of this device component's corresponding Signal process. 
	 * Gets all the properties specified in this device implementation and also its type component. 
	 * For each such property, ASME2SSME_Device_property_call() method is used to make a process call.
	 * 
	 * @param device_impl a device implementation component
	 * @return a ProcessModelImpl (Signal process)
	 */
	public ProcessModelImpl ASME2SSME_Device_property(DeviceImplementation device_impl){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_Device_property: "+ device_impl.getName());

		String p_name = Name_of_Device(aadlColAPI.ComponentImpl_GetName(device_impl));
		String pname = p_name + "_property";
		ProcessModelImpl p_property = myLib.STree_mkSkeletonModel(myLib.mkIdentifier(pname) );

		ArrayList<PropertyAssociation> property_association_list1 = aadlColAPI.ComponentImpl_GetPropertyAssociations(device_impl);
		DeviceType device_type = (DeviceType) aadlColAPI.ComponentImpl_GetComponentType(device_impl);
		ArrayList<PropertyAssociation> property_association_list2 = aadlColAPI.Component_GetPropertyAssociations(device_type);
		property_association_list1.addAll(property_association_list2);
		if(property_association_list1.size()>0){
			for(int i=0; i<property_association_list1.size(); i++)
			{
				PropertyAssociation pa = property_association_list1.get(i);
				ProcessCallImpl pcall1 = ASME2SSME_Device_property_call(pa);
				myLib.STree_addExpressionOfProcessToaModel(p_property, pcall1);   
			}
		}
		return p_property;
	}


	/**
	 * Translates device features
	 * 
	 * @param dt
	 */
	public void ASME2SSME_Device_Features(DeviceType dt, ProcessModelImpl vprocess){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_Device_Features: "+ dt.getName());
		if (dt.getAllFeatures() != null) {
			// PORTS: implemented as a signal in the interface  		
			// EList<DataPort> dport_list = dt.getOwnedDataPorts();
			// EList<EventPort> eport_list = dt.getOwnedEventPorts();
			// EList<EventDataPort> edport_list = dt.getOwnedEventDataPorts();

			//feature group
			ArrayList<FeatureGroup> fgroup_list = aadlColAPI.Component_GetFeatureGroups(dt);
			ASME2SSME_Translate_FeatureGroup asme2ssme_fg = new ASME2SSME_Translate_FeatureGroup();
			for(int i1=0; i1<fgroup_list.size(); i1++)
			{
				FeatureGroup pg = fgroup_list.get(i1);
				if(asme2ssme_fg.ASME2SSME_FeatureGroup_bw(pg) != null){
					SignalsDeclarationImpl in1 = asme2ssme_fg.ASME2SSME_FeatureGroup_bw(pg);
					myLib.STree_addDeclarationToInterface(vprocess, in1, 2);
				}
				if(asme2ssme_fg.ASME2SSME_FeatureGroup_wb(pg) != null){
					SignalsDeclarationImpl out1 = asme2ssme_fg.ASME2SSME_FeatureGroup_wb(pg);
					myLib.STree_addDeclarationToInterface(vprocess, out1, 3);
				}
			}

			//bus access 
			ArrayList<BusAccess> b_list = aadlColAPI.Device_GetBusAccess(dt);
			if(b_list.size()>0){
				for(int i4=0; i4<b_list.size(); i4++)
				{
					BusAccess ba1 = b_list.get(i4);
					ASME2SSME_BusAccess(ba1);			
				}
			}
			//subprogram access: not exist in Version 1.5.8
			//subprogram group access: not exist in version 1.5.8
		}
	}


	/**
	 * Translates flow specifications of a device type
	 * 
	 * @param tt a device type component
	 */
	public void ASME2SSME_Device_FlowSpecs(ProcessModelImpl vprocess, DeviceType dt){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_Device_FlowSpecs: "+ dt.getName());

		if(asme2ssme_common.ASME2SSME_ComponentType_FlowSpecs(dt) != null){
			ProcessModelImpl p1= asme2ssme_common.ASME2SSME_ComponentType_FlowSpecs(dt);

			myLib.STree_addProcessDeclaration(vprocess, p1);
			ArrayList<ExprParam> lparam = new ArrayList<ExprParam> ();
			String pname = Name_of_Flow(aadlColAPI.ComponentType_GetName(dt));
			ProcessExpansionImpl call1 = myLib.STree_mkProcessCallWithParameters(myLib.mkIdentifier(pname), lparam);
			myLib.STree_addExpressionOfProcessToaModel(vprocess, call1);
		}
	}


	/**
	 * Translates flows of a device implementation
	 * 
	 * @param dd a device implementation component
	 */
	public void ASME2SSME_DeviceImpl_Flows(ProcessModelImpl vprocess, DeviceImplementation dd){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_DeviceImpl_Flows: "+ dd.getName());

		asme2ssme_common.ASME2SSME_ComponentImpl_Flows(dd);

		if(asme2ssme_common.ASME2SSME_ComponentImpl_Flows(dd) != null){
			ProcessModelImpl p1= asme2ssme_common.ASME2SSME_ComponentImpl_Flows(dd);

			myLib.STree_addProcessDeclaration(vprocess, p1);
			ArrayList<ExprParam> lparam = new ArrayList<ExprParam> ();
			String pname = Name_of_Flow(dd.getName().replace('.','_'));
			ProcessExpansionImpl call1 = myLib.STree_mkProcessCallWithParameters(myLib.mkIdentifier(pname), lparam);
			myLib.STree_addExpressionOfProcessToaModel(vprocess, call1);
		}
	}


	/**
	 * Translate modes of a device implementation component
	 * 
	 * @param dd
	 */
	public void ASME2SSME_DeviceImpl_Modes(DeviceImplementation dd){

		Trace("\n ====== ASME2SSME_Translate_Device: ASME2SSME_DeviceImpl_Modes: "+ dd.getName());
		asme2ssme_common.ASME2SSME_ComponentImpl_Modes(dd);
	}


}
